/*********************************************************************/
/*                                                                   */
/*             Optimized BLAS libraries                              */
/*                     By Kazushige Goto <kgoto@tacc.utexas.edu>     */
/*                                                                   */
/* Copyright (c) The University of Texas, 2005. All rights reserved. */
/* UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING  */
/* THIS SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF      */
/* MERCHANTABILITY, FITNESS FOR ANY PARTICULAR PURPOSE,              */
/* NON-INFRINGEMENT AND WARRANTIES OF PERFORMANCE, AND ANY WARRANTY  */
/* THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE OF     */
/* TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO   */
/* THE USE OF THE SOFTWARE OR DOCUMENTATION.                         */
/* Under no circumstances shall University be liable for incidental, */
/* special, indirect, direct or consequential damages or loss of     */
/* profits, interruption of business, or related expenses which may  */
/* arise from use of Software or Documentation, including but not    */
/* limited to those resulting from defects in Software and/or        */
/* Documentation, or loss or inaccuracy of data of any kind.         */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include "common.h"

#ifdef SMP_SERVER

/* This is a thread server model implementation.  The threads are   */
/* spawned at first access to blas library, and still remains until */
/* destruction routine is called.  The number of threads are        */
/* equal to "OMP_NUM_THREADS - 1" and thread only wakes up when     */
/* jobs is queued.                                                  */

/* We need this grobal for cheking if initialization is finished.   */
int blas_server_avail = 0;

/* Local Variables */
#if   defined(USE_PTHREAD_LOCK)
static pthread_mutex_t  server_lock    = PTHREAD_MUTEX_INITIALIZER;
#elif defined(USE_PTHREAD_SPINLOCK)
static pthread_spinlock_t  server_lock = 0;
#else
static unsigned long server_lock       = 0;
#endif

static blas_pool_t   pool;
static pthread_t    blas_threads[MAX_CPU_NUMBER];

#undef MONITOR
#define TIMING

#ifdef MONITOR

/* Monitor is a function to see thread's status for every seconds. */
/* Usually it turns off and it's for debugging.                    */

static pthread_t      monitor_thread;
static int main_status[MAX_CPU_NUMBER];
#define MAIN_ENTER	 0x01
#define MAIN_EXIT	 0x02
#define MAIN_TRYLOCK	 0x03
#define MAIN_LOCKSUCCESS 0x04
#define MAIN_QUEUEING	 0x05
#define MAIN_RUNNING     0x06
#define MAIN_WAITING	 0x07
#define MAIN_FINISH      0x08
#endif

#ifdef TIMING
long	exit_time[MAX_CPU_NUMBER];
#endif

#undef TIMING_DEBUG

/* This is a main routine of threads. Each thread waits until job is */
/* queued.                                                           */

static int blas_thread_server(void *arg){

  /* Thread identifier */
  long  cpu = (long)arg;
  void *buffer;
  blas_queue_t	*queue;
#ifdef USE_BUSYWAIT
  int busywaitdriven;
#endif

#ifdef TIMING_DEBUG
  unsigned long start, stop;
#endif

#ifdef MONITOR
    main_status[cpu] = MAIN_ENTER;
#endif

  /* Each server needs each buffer */
  buffer   = blas_memory_alloc(2);

#ifdef SMP_DEBUG
    fprintf(STDERR, "Server[%2ld] Thread is started!\n", cpu);
#endif

  while (1){

#ifdef MONITOR
    main_status[cpu] = MAIN_TRYLOCK;
#endif

#ifdef TIMING
    exit_time[cpu] = rpcc();
#endif

#ifdef USE_BUSYWAIT
    if (!pool.busywait) {
#endif

 /* Waiting for Queue */
      pthread_mutex_lock(&pool.lock);
    
#ifdef MONITOR
      main_status[cpu] = MAIN_LOCKSUCCESS;
#endif
      
      while((pool.queue == NULL)&&(!pool.shutdown)
#ifdef USE_BUSYWAIT
	    && (!pool.busywait)
#endif
	    ){

#ifdef MONITOR
	main_status[cpu] = MAIN_WAITING;
#endif
	pthread_cond_wait(&pool.filled, &pool.lock);
      }
      
#ifdef USE_BUSYWAIT
      if (pool.busywait) {
	pthread_mutex_unlock(&pool.lock);
	continue;
      }
      busywaitdriven = 0;
#endif

      if (pool.shutdown) break;

#ifdef MONITOR
      main_status[cpu] = MAIN_LOCKSUCCESS;
#endif
#ifdef MONITOR
      main_status[cpu] = MAIN_QUEUEING;
#endif
      
      queue = pool.queue;
      pool.queue = queue->next;
      
      if (pool.queue) pthread_cond_signal(&pool.filled);
      pthread_mutex_unlock(&pool.lock);

#ifdef USE_BUSYWAIT
    } else {
      while (pool.busywait && !pool.busyqueue[cpu]) {};
      if (!pool.busywait) continue;
      queue = (blas_queue_t *)pool.busyqueue[cpu];
      busywaitdriven = 1;
    }
#endif

#ifdef MONITOR
    main_status[cpu] = MAIN_RUNNING;
#endif

#ifdef TIMING_DEBUG
    start = rpcc();
#endif

    if (queue) {
      int mode          = queue -> mode;
      long m            = queue -> m;
      long n            = queue -> n;
      long k            = queue -> k;
      void *a           = queue -> a;
      void *b           = queue -> b;
      void *c           = queue -> c;
      long lda          = queue -> lda;
      long ldb          = queue -> ldb;
      long ldc          = queue -> ldc;
      double alpha_r    = queue -> alpha[0];
      double alpha_i    = queue -> alpha[1];
      void *execroutine = queue -> routine;
      long from         = queue -> from;
      long to           = queue -> to;
      void *sa          = queue -> sa;
      void *sb          = queue -> sb;

      
#ifdef SMP_DEBUG
    fprintf(STDERR, "Server[%2ld] Started.  Mode = 0x%03x M = %3ld N=%3ld K=%3ld\n",
	    cpu, queue->mode, queue->m, queue->n, queue->k);
#endif

      /* Now starting calculation.  */
      if (!(mode & BLAS_COMPLEX)){
#ifdef EXPRECISION
	if (mode & BLAS_XDOUBLE){
	  /* REAL / Extended Double */
	  int (*routine)(long, long, long, xdouble, 
			 xdouble *, long, xdouble *, long, 
			 xdouble *, long, void *, void *, long, long);

	  routine = execroutine;

	  if (sa == NULL) 
	    sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);

	  if (sb == NULL)
	    sb = (void *)(((long)sa 
			   + ((XGEMM_P * XGEMM_Q * sizeof(xdouble) + GEMM_ALIGN) & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	  (routine)(m, n, k, (xdouble)alpha_r, a, lda, b, ldb, c, ldc, sa, sb, from, to);
	} else 
#endif
	  if (mode & BLAS_DOUBLE){
	    /* REAL / Double */
	    int (*routine)(long, long, long, double, 
			   double *, long, double *, long,
			   double *, long, void *, void *, long, long);
	    routine = execroutine;

	  if (sa == NULL) 
	    sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);

	  if (sb == NULL)
	    sb = (void *)(((long)sa 
			   + ((DGEMM_P * DGEMM_Q * sizeof(double) + GEMM_ALIGN) & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	    (routine)(m, n, k, (double)alpha_r, a, lda, b, ldb, c, ldc, sa, sb, from ,to);
	  } else {
	    /* REAL / Single */
	    int (*routine)(long, long, long, float, 
			   float *, long, float *, long, 
			   float *, long, void *, void *, long, long);
	    routine = execroutine;

	  if (sa == NULL) 
	    sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);

	  if (sb == NULL)
	    sb = (void *)(((long)sa 
			   + ((SGEMM_P * SGEMM_Q * sizeof(float) + GEMM_ALIGN) & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	    (routine)(m, n, k, (float)alpha_r, a, lda, b, ldb, c, ldc, sa, sb, from, to);
	  }
      } else {
#ifdef EXPRECISION
	if (mode & BLAS_XDOUBLE){
	  /* COMPLEX / Extended Double */
	  int (*routine)(long, long, long, xdouble, xdouble,
			 xdouble *, long, xdouble *, long,
			 xdouble *, long, void *, void *, long, long);
	  routine = execroutine;

	  if (sa == NULL) 
	    sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);

	  if (sb == NULL)
	    sb = (void *)(((long)sa 
			   + ((XGEMM_P * XGEMM_Q * 2 * sizeof(xdouble) + GEMM_ALIGN) & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	  (routine)(m, n, k, (xdouble)alpha_r, (xdouble)alpha_i, a, lda, b, ldb, c, ldc, sa, sb, from, to);
	} else
#endif
	  if (mode & BLAS_DOUBLE){
	    /* COMPLEX / Double */
	    int (*routine)(long, long, long, double, double,
			   double *, long, double *, long,
			   double *, long, void *, void *, long, long);
	    routine = execroutine;

	  if (sa == NULL) 
	    sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);

	  if (sb == NULL)
	    sb = (void *)(((long)sa 
			   + ((ZGEMM_P * ZGEMM_Q * 2 * sizeof(double) + GEMM_ALIGN) & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	    (routine)(m, n, k, (double)alpha_r, (double)alpha_i, a, lda, b, ldb, c, ldc, sa, sb, from, to);
	  } else {
	    /* COMPLEX / Single */
	    int (*routine)(long, long, long, float, float,
			   float *, long, float *, long,
			   float *, long, void *, void *, long, long);
	    routine = execroutine;

	    if (sa == NULL) 
	      sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);
	    
	    if (sb == NULL)
	      sb = (void *)(((long)sa
			     + ((CGEMM_P * CGEMM_Q * 2 * sizeof(float) + GEMM_ALIGN) & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);
	    
	    (routine)(m, n, k, (float)alpha_r, (float)alpha_i, a, lda, b, ldb, c, ldc, sa, sb, from, to);
	  }
      }
    }
   
#ifdef SMP_DEBUG
    fprintf(STDERR, "Server[%2ld] Finished!\n", cpu);
#endif

#ifdef MONITOR
    main_status[cpu] = MAIN_FINISH;
#endif


#ifdef USE_BUSYWAIT
    if (!busywaitdriven) {
#endif
      pthread_mutex_lock(&queue->lock);
      queue -> status = BLAS_STATUS_FINISHED;

      pthread_cond_signal(&queue->finish);
      pthread_mutex_unlock(&queue->lock);

#ifdef USE_BUSYWAIT
    } else {
      pool.busyqueue[cpu] = NULL;
      WMB;
    }
#endif

#ifdef TIMING_DEBUG
    stop = rpcc();

    fprintf(STDERR, "Thread[%ld] : %16lu %16lu (%8lu cycles)  %ld  %ld\n", cpu + 1,
	    start, stop,
	    stop - start, queue->from, queue->to);
#endif
      
  }

  /* Shutdown procedure */

#ifdef SMP_DEBUG
      fprintf(STDERR, "Server[%2ld] Shutdown!\n",  cpu);
#endif

    pthread_mutex_unlock(&pool.lock);

  blas_memory_free(buffer);

  pthread_exit(NULL);

  return 0;
}

#ifdef MONITOR
static int blas_monitor(void *arg){
  int i;

  while(1){
    for (i = 0; i < blas_cpu_number - 1; i++){
      switch (main_status[i]) {
      case MAIN_ENTER :
	fprintf(STDERR, "THREAD[%2d] : Entering.\n", i);
	break;
      case MAIN_EXIT :
	fprintf(STDERR, "THREAD[%2d] : Exiting.\n", i);
	break;
      case MAIN_TRYLOCK :
	fprintf(STDERR, "THREAD[%2d] : Trying lock operation.\n", i);
	break;
      case MAIN_QUEUEING :
	fprintf(STDERR, "THREAD[%2d] : Queueing.\n", i);
	break;
      case MAIN_RUNNING :
	fprintf(STDERR, "THREAD[%2d] : Running.\n", i);
	break;
      case MAIN_WAITING :
	fprintf(STDERR, "THREAD[%2d] : Waiting.\n", i);
	break;
      case MAIN_FINISH :
	fprintf(STDERR, "THREAD[%2d] : Finishing.\n", i);
	break;
      }
    }
    sleep(1);
  }
  
 return 0;
}
#endif

#undef NEED_STACKATTR

/* Initializing routine */
int blas_thread_init(void){
  long i;
#ifdef NEED_STACKATTR
  pthread_attr_t attr;
#endif

#ifndef ALL_THREADED
  if (blas_server_avail || (blas_cpu_number <= 1)) return 0;
#else
  if (blas_server_avail) return 0;
#endif
  
#ifdef NEED_STACKATTR
  pthread_attr_init(&attr);
  pthread_attr_setguardsize(&attr, 0);
  pthread_attr_setstacksize( &attr, 0x4000U);
#endif

  LOCK_COMMAND(&server_lock);

  if (!blas_server_avail){

    blas_pool_init(&pool);

#ifdef ALL_THREADED
    for(i = 0; i < blas_cpu_number; i++){
#else
    for(i = 0; i < blas_cpu_number - 1; i++){
#endif

#ifdef NEED_STACKATTR
      pthread_create(&blas_threads[i], &attr, 
		     (void *)&blas_thread_server, (void *)i);
#else
      pthread_create(&blas_threads[i], NULL, 
		     (void *)&blas_thread_server, (void *)i);
#endif
    }

#ifdef MONITOR
    pthread_create(&monitor_thread, NULL, 
		     (void *)&blas_monitor, (void *)NULL);
#endif

    blas_server_avail = 1;
  }

  UNLOCK_COMMAND(&server_lock);

  return 0;
}

/* 
   User can call one of two routines.

     exec_blas_async ... immediately returns after jobs are queued.

     exec_blas       ... returns after jobs are finished.
*/

int exec_blas_async(blas_queue_t *queue){

  blas_queue_t *current;

#ifdef USE_BUSYWAIT
  int i;

  if (!pool.busywait) {
#endif
    pthread_mutex_lock(&pool.lock);

    /* pool.queue should be here (by Brian Summer) */
    if (pool.queue) {
      current = pool.queue;
      while (current -> next) current = current -> next;
      current -> next = queue;
    } else {
      pool.queue = queue;
  }
    
    pthread_cond_signal(&pool.filled);
    pthread_mutex_unlock(&pool.lock);

#ifdef USE_BUSYWAIT
  } else {
    i = 0;
    while (queue) {
      if (!pool.busyqueue[i]){
	pool.busyqueue[i] = queue;
	queue = queue -> next;
      }
      i ++;
      if (i >= blas_cpu_number) i = 0;
    }
  }
#endif

  return 0;
}

int exec_blas_async_wait(blas_queue_t *queue){

#ifdef USE_BUSYWAIT
  int i;
#endif

#ifdef SMP_DEBUG
    fprintf(STDERR, "Syncing...");
#endif

#ifdef USE_BUSYWAIT
  if (!pool.busywait) {
#endif
    while (queue){
      pthread_mutex_lock(&queue->lock);
      while(queue->status != BLAS_STATUS_FINISHED)
	pthread_cond_wait(&queue->finish, &queue->lock);
      pthread_mutex_unlock(&queue->lock);
      
      if (queue -> last) break;
      queue = queue -> next;
    }
#ifdef USE_BUSYWAIT
  } else {
    for (i = 0; i < blas_cpu_number; i++) {
      while (pool.busyqueue[i]) {};
    }
  }
#endif

#ifdef SMP_DEBUG
    fprintf(STDERR, "Done.\n\n");
#endif

  return 0;
}

/* Execute Threads */
int exec_blas(blas_queue_t *queue, void *buffer){

#ifndef ALL_THREADED
  int exec_flag;
  void *sa, *sb;
#endif
#ifdef TIMING_DEBUG
  unsigned long start, stop;
#endif

  if (queue == NULL) return 0;

#ifndef ALL_THREADED
  exec_flag = 0;

  if (queue -> next) {
    exec_blas_async(queue -> next);
    exec_flag = 1;
  }
  
#ifdef TIMING_DEBUG
  start = rpcc();

  fprintf(STDERR, "\n");
#endif

 {
    if (!(queue -> mode & BLAS_COMPLEX)){ /* REAL */
#ifdef EXPRECISION
      if (queue -> mode & BLAS_XDOUBLE){
	int (*routine)(long, long, long, xdouble, 
		       xdouble *, long, xdouble *, long,
		       xdouble *, long, void *, void *, long, long);
	routine = queue -> routine;

	sa = queue -> sa;
	sb = queue -> sb;

	if (sa == NULL) 
	  sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);
	
	if (sb == NULL)
	  sb = (void *)(((long)sa 
			 + ((XGEMM_P * XGEMM_Q * sizeof(xdouble) + GEMM_ALIGN)
			    & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	(routine)(queue -> m, queue -> n, queue -> k,
		  (xdouble)queue -> alpha[0],
		  queue -> a, queue -> lda,
		  queue -> b, queue -> ldb,
		  queue -> c, queue -> ldc,
		  sa, sb,
		  queue -> from, queue -> to);
      } else
#endif
	if (queue -> mode & BLAS_DOUBLE){
	  int (*routine)(long, long, long, double, 
			 double *, long, double *, long,
			 double *, long, void *, void *, long, long);
	  routine = queue -> routine;
	  
	  sa = queue -> sa;
	  sb = queue -> sb;
	  
	  if (sa == NULL) 
	    sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);
	  
	  if (sb == NULL)
	    sb = (void *)(((long)sa 
			   + ((DGEMM_P * DGEMM_Q * sizeof(double) + GEMM_ALIGN)
			      & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);
	  
	  (routine)(queue -> m, queue -> n, queue -> k,
		    (double)queue -> alpha[0],
		    queue -> a, queue -> lda,
		    queue -> b, queue -> ldb,
		    queue -> c, queue -> ldc,
		    sa, sb,
		    queue -> from, queue -> to);
	} else {
	  int (*routine)(long, long, long, float, 
			 float *, long, float *, long,
			 float *, long, void *, void *, long, long);
	  routine = queue -> routine;

	  sa = queue -> sa;
	  sb = queue -> sb;
	  
	  if (sa == NULL) 
	    sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);
	  
	  if (sb == NULL)
	    sb = (void *)(((long)sa 
			   + ((SGEMM_P * SGEMM_Q * sizeof(float) + GEMM_ALIGN)
			      & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	  (routine)(queue -> m, queue -> n, queue -> k,
		    (float)queue -> alpha[0],
		    queue -> a, queue -> lda,
		    queue -> b, queue -> ldb,
		    queue -> c, queue -> ldc,
		    sa, sb,
		    queue -> from, queue -> to);
      }
    } else {
#ifdef EXPRECISION
      if (queue -> mode & BLAS_XDOUBLE){
	int (*routine)(long, long, long, xdouble, xdouble,
		       xdouble *, long, xdouble *, long,
		       xdouble *, long, void *, void *, long, long);
	routine = queue -> routine;

	sa = queue -> sa;
	sb = queue -> sb;

	if (sa == NULL) 
	  sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);
	
	if (sb == NULL)
	  sb = (void *)(((long)sa 
			 + ((XGEMM_P * XGEMM_Q * 2 * sizeof(xdouble) + GEMM_ALIGN)
			    & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	(routine)(queue -> m, queue -> n, queue -> k,
		  (xdouble)queue -> alpha[0], (xdouble)queue -> alpha[1],
		  queue -> a, queue -> lda,
		  queue -> b, queue -> ldb,
		  queue -> c, queue -> ldc,
		  sa, sb,
		  queue -> from, queue -> to);
      } else
#endif
      if (queue -> mode & BLAS_DOUBLE){
	int (*routine)(long, long, long, double, double,
		       double *, long, double *, long,
		       double *, long, void *, void *, long, long);
	routine = queue -> routine;

	sa = queue -> sa;
	sb = queue -> sb;

	if (sa == NULL) 
	  sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);
	
	if (sb == NULL)
	  sb = (void *)(((long)sa 
			 + ((ZGEMM_P * ZGEMM_Q * 2 * sizeof(double) + GEMM_ALIGN)
			    & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	(routine)(queue -> m, queue -> n, queue -> k,
		  (double)queue -> alpha[0], (double)queue -> alpha[1],
		  queue -> a, queue -> lda,
		  queue -> b, queue -> ldb,
		  queue -> c, queue -> ldc,
		  sa, sb,
		  queue -> from, queue -> to);
      } else {
	int (*routine)(long, long, long, float, float,
		       float *, long, float *, long,
		       float *, long, void *, void *, long, long);
	routine = queue -> routine;

	sa = queue -> sa;
	sb = queue -> sb;

	if (sa == NULL) 
	  sa = (void *)((long)buffer + GEMM_BUFFER_A_OFFSET);
	
	if (sb == NULL)
	  sb = (void *)(((long)sa 
			 + ((CGEMM_P * CGEMM_Q * 2 * sizeof(float) + GEMM_ALIGN)
			    & ~GEMM_ALIGN)) + GEMM_BUFFER_B_OFFSET);

	(routine)(queue -> m, queue -> n, queue -> k,
		  (float)queue -> alpha[0], (float)queue -> alpha[1],
		  queue -> a, queue -> lda,
		  queue -> b, queue -> ldb,
		  queue -> c, queue -> ldc,
		  sa, sb,
		  queue -> from, queue -> to);
      }
    }
  }

#ifdef TIMING_DEBUG
    stop = rpcc();
#endif

  if (exec_flag) exec_blas_async_wait(queue -> next);

#else
  exec_blas_async(queue);
  exec_blas_async_wait(queue);
#endif

#ifdef TIMING_DEBUG
    fprintf(STDERR, "Thread[0] : %16lu %16lu (%8lu cycles)  %ld   %ld\n", 
	    start, stop,
	    stop - start, queue->from, queue->to);
#endif

  return 0;
}

/* Shutdown procedure, but user don't have to call this routine. The */
/* kernel automatically kill threads.                                */

int blas_thread_shutdown_(void){

  int i;

  if (!blas_server_avail) return 0;
  
  LOCK_COMMAND(&server_lock);

  if (blas_server_avail){
    pthread_mutex_lock(&pool.lock);
    pool.shutdown = 1;

    pthread_cond_broadcast(&pool.filled);
    pthread_mutex_unlock(&pool.lock);
    
    for(i = 0; i < blas_cpu_number - 1; i++){
      pthread_join(blas_threads[i], NULL);
    }

    blas_server_avail = 0;
  }
  
  UNLOCK_COMMAND(&server_lock);
  
  return 0;
}

#ifdef USE_BUSYWAIT
int blas_busywait(int flag) {

  int oldflag;

  pthread_mutex_lock(&pool.lock);

  oldflag = pool.busywait;
  pool.busywait = flag;
  
  if (!oldflag && flag) pthread_cond_broadcast(&pool.filled);
  pthread_mutex_unlock(&pool.lock);

  return 0;
}

int BLASFUNC(blas_busywait_on_)(void) {

  blas_busywait(1);

  return 0;
}

int BLASFUNC(blas_busywait_off_)(void) {

  blas_busywait(0);

  return 0;
}

#endif
#endif

